from numpy import random
from tabulate import tabulate
from loguru import logger
from pogema_toolbox.views.view_utils import View, eval_logs_to_pandas, drop_na
from typing import Literal


class TabularView(View):
    print_results: bool = False
    type: Literal['tabular'] = 'tabular'
    table_format: str = 'psql'
    show_std: bool = True
    skip_zero_std: bool = True


def preprocess_table(eval_configs, view):
    df = eval_logs_to_pandas(eval_configs)

    drop_na(df)

    # Ensure eval_configs[0]['metrics'] and view.drop_keys are lists or convert them
    metrics_keys = [key for key in eval_configs[0].get('metrics', []) if key in df.columns]
    drop_keys = list(view.drop_keys) if hasattr(view, 'drop_keys') else []
    group_by = [col for col in df.columns if col not in metrics_keys + drop_keys]

    # Define aggregation operations
    agg_ops = {key: ['mean', 'std'] for key in metrics_keys} if view.show_std else {key: 'mean' for key in metrics_keys}

    # Perform aggregation
    df_agg = df.groupby(by=group_by, as_index=False).agg(agg_ops).round(view.round_digits)

    if view.show_std:
        df_agg.columns = ['_'.join(col).rstrip('_') for col in df_agg.columns]

        for key in metrics_keys:
            mean_col = f"{key}_mean"
            std_col = f"{key}_std"
            valid_std = (df_agg[std_col].notna()) & ((df_agg[std_col] != 0) | view.skip_zero_std)

            # Create a combined mean and std string for valid std values
            combined_str = df_agg[mean_col].astype(str) + " ± " + df_agg[std_col].astype(str)
            df_agg[key] = df_agg[mean_col].astype(str)  # Initialize the key column with mean values as strings

            # Update the key column with combined_str where valid_std is True
            df_agg.loc[valid_std, key] = combined_str

            # Drop the now redundant mean and std columns
            df_agg.drop(columns=[mean_col, std_col], inplace=True)

    # Drop specified keys, handle sorting, and renaming if required
    df_agg.drop(columns=drop_keys, errors='ignore', inplace=True)
    if hasattr(view, 'sort_by') and view.sort_by:
        df_agg.sort_values(by=view.sort_by, inplace=True)
    if hasattr(view, 'rename_fields') and view.rename_fields:
        df_agg.rename(columns=view.rename_fields, inplace=True)

    return df_agg


def process_table_view(results, view: TabularView):
    df = preprocess_table(results, view)

    table = tabulate(df, headers='keys', tablefmt=view.table_format)
    if view.print_results:
        logger.info('\n' + table)


def generate_mock_data(num_results=25, ):
    results = []
    for i in range(num_results):
        result = {
            "metrics": {
                "avg_throughput": random.uniform(0.01, 0.25),  # Random float between 0.01 and 0.25
                "avg_num_agents_in_obs": random.uniform(1.4, 1.7),  # Random float between 1.4 and 1.7
                "runtime": random.uniform(29, 55)  # Random float between 29 and 55
            },
            "env_grid_search": {
                "map_name": f"mazes-seed-{i}-10x10",
                "num_agents": random.randint(10, 16),
                "seed": random.randint(0, 3)  # Random int between 0 and 1
            },
            "algorithm": "MATS-LP"
        }
        results.append(result)
    return results


def main():
    results = generate_mock_data()
    process_table_view(results, TabularView(round_digits=2, print_results=True, drop_keys=['seed', 'map_name']))
    process_table_view(results, TabularView(round_digits=2, print_results=True, drop_keys=[]))


if __name__ == '__main__':
    main()



from typing_extensions import Literal
from typing import Optional

from pydantic import BaseModel


class AlgoBase(BaseModel):
    name: str = None
    num_process: int = 3
    device: str = 'cuda'
    parallel_backend: Literal['dask', 'sequential', 'balanced_dask', 'balanced_dask'] = 'balanced_dask'
    seed: Optional[int] = 0
    preprocessing: Optional[str] = None



import seaborn as sns

import numpy as np
from matplotlib import pyplot as plt
from typing import Optional

from loguru import logger
from pogema_toolbox.views.view_utils import View, eval_logs_to_pandas, drop_na
from typing import Tuple

from typing_extensions import Literal


def custom_palette():
    q = list(sns.color_palette("deep"))
    q[1], q[9] = q[9], q[1]
    q[5], q[9] = q[9], q[5]
    q[6], q[9] = q[9], q[6]
    return q


class PlotView(View):
    type: Literal['plot'] = 'plot'
    name: str = None
    x: str = None
    y: str = None
    by: str = 'algorithm'
    width: float = 2.6
    height: float = 2.8
    remove_title: bool = False
    line_width: float = 2.0

    error_bar: Tuple[str, int] = ('ci', 95)

    plt_style: str = 'seaborn-v0_8-colorblind'
    figure_dpi: int = 300
    font_size: int = 8
    legend_font_size: int = 7
    legend_loc: Literal[
        'best', 'upper right', 'upper left', 'lower left', 'lower right', 'right',
        'center left', 'center right', 'lower center', 'upper center', 'center'] = 'best'
    figure_face_color: str = '#FFFFFF'
    use_log_scale_x: bool = False
    use_log_scale_y: bool = False
    markers: bool = True
    line_types: bool = True
    extension: Literal['svg', 'png', 'pdf'] = 'pdf'
    palette: list = custom_palette()
    hue_order: list = None

    tight_layout: bool = True
    ticks: Optional[list] = None
    remove_legend_title: bool = True


def process_plot_view(results, view: PlotView, save_path=None):
    df = eval_logs_to_pandas(results)
    df = drop_na(df)
    if view.hue_order is None:
        view.hue_order = sorted(df['algorithm'].unique())

    if view.sort_by:
        df.sort_values(by=['map_name', 'algorithm'], inplace=True)

    if view.rename_fields:
        df = df.rename(columns=view.rename_fields)

    prepare_plt(view)
    x, y, hue = prepare_plot_fields(view)

    fig, ax = plt.subplots()
    if x not in df.keys():
        logger.warning(f"Could not interpret value {x} for parameter 'x'. Skipping this plot.")
        return
    if y not in df.keys():
        logger.warning(f"Could not interpret value {y} for parameter 'y'. Skipping this plot.")
        return

    sns.lineplot(x=x, y=y, data=df, errorbar=view.error_bar, hue=hue, hue_order=view.hue_order,
                 style_order=view.hue_order, linewidth=view.line_width,
                 style=hue if view.line_types else None, markers=view.markers,
                 palette=view.palette[:len(view.hue_order)], ax=ax,
                 )
    if not view.remove_title:
        ax.set_title(view.name)

    if view.remove_legend_title:
        ax.legend().set_title('')

    if view.use_log_scale_x:
        ax.set_xscale('log', base=2)
        from matplotlib.ticker import ScalarFormatter
        ax.xaxis.set_major_formatter(ScalarFormatter())

    if view.use_log_scale_y:
        ax.set_yscale('log', base=2)
        from matplotlib.ticker import ScalarFormatter
        ax.yaxis.set_major_formatter(ScalarFormatter())

    plt.tight_layout()
    plt.grid()

    if view.ticks:
        ax.set_xticks(np.array(view.ticks))

    plt.plot()

    if save_path:
        plt.savefig(save_path)

    plt.close()


def prepare_plt(view: PlotView):
    plt.style.use(view.plt_style)
    plt.rcParams['figure.figsize'] = (view.width, view.height)
    plt.rcParams['figure.dpi'] = view.figure_dpi
    plt.rcParams['font.size'] = view.font_size
    plt.rcParams['legend.fontsize'] = view.legend_font_size
    plt.rcParams['legend.loc'] = view.legend_loc
    plt.rcParams['figure.facecolor'] = view.figure_face_color

    if view.name:
        plt.title(view.name)


def prepare_plot_fields(view):
    x = view.x if view.x not in view.rename_fields else view.rename_fields[view.x]
    y = view.y if view.y not in view.rename_fields else view.rename_fields[view.y]
    hue = view.by if view.by not in view.rename_fields else view.rename_fields[view.by]
    return x, y, hue



from typing import Union, List

import pandas as pd
import yaml
from pydantic import BaseModel
from collections import defaultdict
from tabulate import tabulate

import json
from loguru import logger


class View(BaseModel):
    drop_keys: list = ['seed']
    round_digits: int = 2
    rename_fields: dict = {"algorithm": "Algorithm",
                           "num_agents": "Number of Agents",
                           "avg_throughput": "Average Throughput",
                           "runtime": "Runtime (seconds)",
                           "max_episode_steps": "Episode Length"
                           }
    rename_algorithms: dict = {}
    sort_by: Union[str, List[str]] = None


def drop_na(df):
    # Drop columns with any NaN values
    cols_before_drop_na = set(df.columns)
    df = df.dropna(axis=1, how='any')
    cols_after_drop_na = set(df.columns)

    # Log dropped columns
    dropped_cols = cols_before_drop_na - cols_after_drop_na
    for col in dropped_cols:
        logger.warning(f"Column '{col}' dropped due to missing values")

    return df


def eval_logs_to_pandas(eval_configs):
    data = {}
    for idx, config in enumerate(eval_configs):
        data[idx] = {**config['env_grid_search'], 'algorithm': config['algorithm']}

        # Adding metrics separately to skip possible lists of metrics (e.g. every step throughput)
        for key, value in config['metrics'].items():
            if isinstance(value, list):
                continue
            data[idx][key] = value
    return pd.DataFrame.from_dict(data, orient='index')


def load_from_folder(folder_path):
    eval_config_path = folder_path / (folder_path.name + '.yaml')
    # check if the file with name current_dir_name.name + '.yaml' exists
    assert eval_config_path.exists(), f'Config file {eval_config_path} does not exist'

    with open(eval_config_path, 'r') as f:
        evaluation_config = yaml.safe_load(f)

    results = []
    # Load results from *.json files
    for file in folder_path.glob('*.json'):
        with open(file, 'r') as f:
            results += json.load(f)
    return results, evaluation_config


def check_seeds(results):
    seed_data = defaultdict(lambda: defaultdict(list))
    algorithms = set()
    problem_count = 0  # Initialize problem counter

    # Populate seed_data and algorithms set
    for res in results:
        if 'env_grid_search' not in res:
            logger.error("env_grid_search data missing")
            return

        env_data = res.get('env_grid_search', {})

        if 'map_name' not in env_data:
            logger.error("No map_name in env_grid_search data")
            return

        if 'num_agents' not in env_data:
            logger.error("No num_agents in env_grid_search data")
            return

        if 'seed' not in env_data:
            logger.error("No seed in env_grid_search data")
            return

        map_name = res['env_grid_search']['map_name']
        num_agents = res['env_grid_search']['num_agents']
        seed = res['env_grid_search']['seed']
        algo = res['algorithm']
        algorithms.add(algo)
        seed_data[(map_name, num_agents, seed)][algo].append(res)

    # Prepare data for tabulate
    tabulate_data = []
    headers = ['Map Name', 'Num Agents', 'Seed'] + sorted(algorithms)

    for (map_name, num_agents, seed), algos in seed_data.items():
        row_issues = False
        row = [map_name, num_agents, seed]
        for algo in sorted(algorithms):
            if algo in algos:
                if len(algos[algo]) > 1:  # Duplicated seeds for this algo
                    row.append(f"x{len(algos[algo])}")
                    problem_count += 1  # Increment problem counter
                    row_issues = True
                else:
                    row.append("ok")  # Seed present and not duplicated
            else:
                row.append("missing")  # Seed missing for this algo
                problem_count += 1  # Increment problem counter
                row_issues = True

        if row_issues:  # Add row only if there are issues
            tabulate_data.append(row)

    # If problems detected, show error with tabulate
    if problem_count > 0:
        table = tabulate(tabulate_data, headers=headers, tablefmt='psql')
        logger.error(f"Detected {problem_count} problems with seeds:\n" + table)
    else:
        logger.success(f"Passed seeds consistency check")



import re
from copy import deepcopy

import numpy as np
from loguru import logger
from pogema import pogema_v0, GridConfig
from pogema.animation import AnimationMonitor, AnimationConfig
from gymnasium import Wrapper
from pydantic import BaseModel

from pogema_toolbox.registry import ToolboxRegistry


class Environment(GridConfig, ):
    with_animation: bool = False
    use_maps: bool = True


class ProvideGlobalObstacles(Wrapper):
    def get_global_obstacles(self):
        return self.grid.get_obstacles().astype(int).tolist()

    def get_global_agents_xy(self):
        return self.grid.get_agents_xy()

    def get_global_targets_xy(self):
        return self.grid.get_targets_xy()


class MultiMapWrapper(Wrapper):
    def __init__(self, env):
        super().__init__(env)
        self._configs = []
        self._rnd = np.random.default_rng(self.grid_config.seed)
        pattern = self.grid_config.map_name

        if pattern:
            maps = ToolboxRegistry.get_maps()
            for map_name in sorted(maps):
                if re.match(pattern, map_name):
                    cfg = deepcopy(self.grid_config)
                    cfg.map = maps[map_name]
                    cfg.map_name = map_name
                    cfg = GridConfig(**cfg.dict())
                    self._configs.append(cfg)
            if not self._configs:
                raise KeyError(f"No map matching: {pattern}")

    def reset(self, seed=None, **kwargs):
        if seed is None:
            seed = self.grid_config.seed
        self._rnd = np.random.default_rng(seed)
        if self._configs is not None and len(self._configs) >= 1:
            map_idx = self._rnd.integers(0, len(self._configs))
            cfg = deepcopy(self._configs[map_idx])
            self.env.unwrapped.grid_config = cfg
            self.env.unwrapped.grid_config.seed = seed
        return self.env.reset(seed=seed, **kwargs)


def create_env_base(config: Environment):
    env = pogema_v0(grid_config=config)
    env = ProvideGlobalObstacles(env)
    if config.use_maps:
        env = MultiMapWrapper(env)
    if config.with_animation:
        env = AnimationMonitor(env, AnimationConfig(directory='experiments/renders', save_every_idx_episode=None))

    return env



# noinspection PyUnresolvedReferences
from pogema_toolbox import fix_num_threads_issue

import json
from pathlib import Path
from loguru import logger
import time

import numpy as np

from pogema_toolbox.config_variant_generator import generate_variants
from pogema_toolbox.create_env import Environment
from pogema_toolbox.registry import ToolboxRegistry

from pogema_toolbox.views.view_multi_plot import process_multi_plot_view, MultiPlotView
from pogema_toolbox.views.view_plot import process_plot_view, PlotView
from pogema_toolbox.views.view_tabular import process_table_view, TabularView
from pogema_toolbox.results_holder import ResultsHolder


def run_episode(env, algo):
    """
    Runs an episode in the environment using the given algorithm.

    Args:
        env: The environment to run the episode in.
        algo: The algorithm used for action selection.

    Returns:
        ResultsHolder: Object containing the results of the episode.
    """
    algo.reset_states()
    results_holder = ResultsHolder()
    # todo provide seed as a parameter
    obs, _ = env.reset(seed=env.grid_config.seed)
    while True:
        obs, rew, terminated, truncated, infos = env.step(algo.act(obs))
        results_holder.after_step(infos)

        if all(terminated) or all(truncated):
            break
    return results_holder.get_final()


def sequential_backend(algo_config, env_configs, full_algo_name, registry_state=None):
    """
    Runs the algorithm sequentially on multiple environments.

    Args:
        algo_config: Configuration for the algorithm.
        env_configs: List of environment configurations.
        full_algo_name: Full name of the algorithm.
        registry_state:

    Returns:
        List: Results of running the algorithm on the environments.
    """
    registry = ToolboxRegistry
    if registry_state is not None:
        registry.recreate_from_state(registry_state)

    results = []
    algo_name = algo_config['name']
    algo = registry.create_algorithm(algo_name, **algo_config)
    algo_cfg = registry.create_algorithm_config(algo_name, **algo_config)
    for idx, env_config in enumerate(env_configs):
        logger.info(f'Running: {full_algo_name} [{idx + 1}/{len(env_configs)}]')
        env = registry.create_env(env_config['name'], **env_config)
        if algo_cfg.preprocessing:
            logger.debug('Adding preprocessing')
            env = registry.create_algorithm_preprocessing(env, algo_name, **algo_config)
        results.append(run_episode(env, algo))

        if env_config.get('with_animation', None):
            from pathlib import Path

            directory = Path(f'renders/{full_algo_name}/')
            name = env.pick_name(env.grid_config)

            directory.mkdir(parents=True, exist_ok=True)
            logger.info(f'Saving animation to "{directory / name}"')
            env.save_animation(name=directory / name)
    return results


def split_on_chunks(size, num_chunks):
    """
    Splits the given size into equal chunks.

    Args:
        size: The size to be split.
        num_chunks: Number of chunks to split into.

    Yields:
        Tuple[int, int]: Start and end indexes of each chunk.
    """
    offset = int(1.0 * size / num_chunks + 0.5)
    for i in range(0, num_chunks - 1):
        yield i * offset, i * offset + offset
    yield num_chunks * offset - offset, size


def get_env_config_cost(raw_config):
    gc = Environment(**raw_config)
    return gc.num_agents * gc.max_episode_steps


def get_balanced_buckets_indexes(env_configs, num_buckets):
    """
    Distributes environment indexes into balanced buckets based on their costs.

    Args:
        env_configs: List of environment configurations.
        num_buckets: Number of buckets to distribute the indexes into.

    Returns:
        List[List[int]]: Balanced buckets containing environment indexes.
    """
    buckets = [[] for _ in range(num_buckets)]
    bucket_costs = [0 for _ in range(num_buckets)]
    env_costs = [get_env_config_cost(ec) for ec in env_configs]
    indexes = np.argsort(env_costs)[::-1]

    for idx in indexes:
        min_bucket_idx = np.argmin(bucket_costs)
        buckets[min_bucket_idx].append(idx)
        bucket_costs[min_bucket_idx] += env_costs[idx]

    # remove empty buckets
    buckets = [bucket for bucket in buckets if len(bucket) > 0]

    return buckets


def get_num_of_available_cpus():
    """
    Returns the number of available CPUs.

    Returns:
        int: Number of available CPUs.
    """
    import multiprocessing
    return multiprocessing.cpu_count()


def dask_backend(algo_config, env_configs, full_algo_name):
    """
    Runs the algorithm using Dask for distributed computing.

    Args:
        algo_config: Configuration for the algorithm.
        env_configs: List of environment configurations.
        full_algo_name: Full name of the algorithm.

    Returns:
        List: Results of running the algorithm on the environments.
    """
    import dask.distributed as dd
    initialized_algo_config = ToolboxRegistry.create_algorithm_config(algo_config['name'], **algo_config)

    num_process = min(initialized_algo_config.num_process, get_num_of_available_cpus())
    cluster = dd.LocalCluster(n_workers=num_process, threads_per_worker=1, nthreads=1, )
    client = dd.Client(cluster, timeout="120s")  # Connect the client to the cluster

    futures = []
    for left, right in split_on_chunks(len(env_configs), initialized_algo_config.num_process):
        future = client.submit(sequential_backend, algo_config, env_configs[left:right], full_algo_name, pure=False)
        futures.append(future)

    results = client.gather(futures)  # Gather the results from the distributed tasks
    client.close()  # Close the Dask client and cluster
    cluster.close()
    results = np.concatenate(results).tolist()
    return results


def balanced_dask_backend(algo_config, env_configs, full_algo_name):
    """
    Runs the algorithm in a balanced manner using Dask for distributed computing.

    Args:
        algo_config: Configuration for the algorithm.
        env_configs: List of environment configurations.
        full_algo_name: Full name of the algorithm.

    Returns:
        List: Results of running the algorithm on the environments.
    """
    logger.info('Running experiment with balanced task backend')
    import dask.distributed as dd

    initialized_algo_config = ToolboxRegistry.create_algorithm_config(algo_config['name'], **algo_config)

    num_process = min(initialized_algo_config.num_process, get_num_of_available_cpus())
    balanced_buckets = get_balanced_buckets_indexes(env_configs, num_process)

    cluster = dd.LocalCluster(n_workers=num_process, threads_per_worker=1, nthreads=1)
    client = dd.Client(cluster, timeout="120s")  # Connect the client to the cluster

    futures = []

    # Getting maps to initialize registry (if not) and  avoid multiple loading
    ToolboxRegistry.get_maps()
    registry_state = ToolboxRegistry.get_state()

    for bucket in balanced_buckets:
        bucket_configs = [env_configs[idx] for idx in bucket]
        future = client.submit(sequential_backend, algo_config, bucket_configs, full_algo_name, registry_state, pure=False)
        futures.append(future)

    results = client.gather(futures)
    client.close()
    cluster.close()

    # Reorder the results according to the original order of env_configs
    ordered_results = [None for _ in range(len(env_configs))]
    for idx, bucket in enumerate(balanced_buckets):
        bucket_results = results[idx]
        for i, env_idx in enumerate(bucket):
            ordered_results[env_idx] = bucket_results[i]

    return ordered_results


def join_metrics_and_configs(metrics, evaluation_configs, env_grid_search, algo_config, algo_name):
    """
    Joins metrics, evaluation configurations, environment grid search, and algorithm name into a result dictionary.

    Args:
        metrics: List of metrics.
        evaluation_configs: List of evaluation configurations.
        env_grid_search: List of environment grid search configurations.
        algo_config: Configuration for the algorithm.
        algo_name: Name of the algorithm.

    Returns:
        List[dict]: List of result dictionaries.
    """
    env_grid_search = [{key[-1]: value for key, value in x.items()} for x in env_grid_search]
    results = []
    for idx, metric in enumerate(metrics):
        results.append({'metrics': metrics[idx], 'env_grid_search': env_grid_search[idx], 'algorithm': algo_name})
    return results


def run_views(results, evaluation_config, eval_dir=None):
    """
    Runs the views specified in the evaluation configuration on the results.

    Args:
        results: List of result dictionaries.
        evaluation_config: Configuration for the evaluation.
        eval_dir: Directory to save the views (optional).

    Returns:
        List: Results of running the views.
    """
    view_results = []
    if 'results_views' not in evaluation_config:
        logger.info("No result views provided in config")
        return
    for key, view in evaluation_config['results_views'].items():
        save_path = Path(eval_dir if eval_dir else '.') / f'{key}.pdf'
        # create directory if not exists
        save_path.parent.mkdir(parents=True, exist_ok=True)

        if view['type'] == 'tabular':
            view_results.append(process_table_view(results, TabularView(**view)))
        elif view['type'] == 'plot':
            view_results.append(process_plot_view(results, PlotView(**view), save_path))
        elif view['type'] == 'multi-plot':
            view_results.append(process_multi_plot_view(results, MultiPlotView(**view), save_path))
    return view_results


def evaluation(evaluation_config, eval_dir=None):
    """
    Runs the evaluation based on the evaluation configuration.

    Args:
        evaluation_config: Configuration for the evaluation.
        eval_dir: Directory to save the evaluation results (optional).

    Returns:
        List: Results of the evaluation.
    """
    env_grid_search, environment_configs = zip(*generate_variants(evaluation_config['environment']))

    results = []
    for key, algo_cfg in evaluation_config['algorithms'].items():
        p_algo_cfg = ToolboxRegistry.create_algorithm_config(algo_cfg['name'], **algo_cfg)
        logger.info(f'Starting: {key}, {algo_cfg}')
        start_time = time.monotonic()
        if p_algo_cfg.parallel_backend == 'sequential':
            metrics = sequential_backend(algo_cfg, environment_configs, key)
        # elif p_algo_cfg.parallel_backend == 'multiprocessing':
        #     metrics = multiprocess_backend(algo_cfg, environment_configs, key)
        elif p_algo_cfg.parallel_backend == 'dask':
            metrics = dask_backend(algo_cfg, environment_configs, key)
        # elif p_algo_cfg.parallel_backend == 'balanced_multiprocessing':
        #     metrics = balanced_multiprocess_backend(algo_cfg, environment_configs, key)
        elif p_algo_cfg.parallel_backend == 'balanced_dask':
            metrics = balanced_dask_backend(algo_cfg, environment_configs, key)
        else:
            raise ValueError(f'Unknown parallel backend: {p_algo_cfg.parallel_backend}')
        algo_results = join_metrics_and_configs(metrics, environment_configs, env_grid_search, algo_cfg, key)
        if eval_dir:
            save_path = Path(eval_dir) / f'{key}.json'
            save_path.parent.mkdir(parents=True, exist_ok=True)
            with open(save_path, "w") as f:
                json.dump(algo_results, f)
        results += algo_results
        logger.success(f'Finished: {key}, runtime: {time.monotonic() - start_time}')

    run_views(results, evaluation_config, eval_dir=eval_dir)
    return results



import codecs
import os
import re

from setuptools import setup, find_packages

cur_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cur_dir, 'README.md'), 'rb') as f:
    lines = [x.decode('utf-8') for x in f.readlines()]
    lines = ''.join([re.sub('^<.*>\n$', '', x) for x in lines])
    long_description = lines


def read(*parts):
    with codecs.open(os.path.join(cur_dir, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


setup(
    name='pogema-toolbox',
    author='Alexey Skrynnik',
    license='Apache License 2.0',
    version=find_version("pogema_toolbox", "__init__.py"),
    description='Evaluation toolbox for Pogema environment',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Tviskaron/pogema-toolbox',
    install_requires=[
        "loguru<=0.7.2",
        "wandb>=0.12.9,<=0.13.4",
        "matplotlib<=3.8.3",
        "seaborn~=0.13.2",
        "tabulate>=0.8.7,<=0.8.10",
        "importlib-metadata==4.13.0",
        "dask[distributed]",
        "pydantic>=1.8.2,<=1.9.1",
        "numpy>=1.21",
        "pandas<=2.2.1",
        "PyYAML<=6.0.1",
        "pogema>=1.2.2"
    ],
    extras_require={

    },
    package_dir={'': './'},
    packages=find_packages(where='./', include='pogema_toolbox*'),
    include_package_data=True,
    python_requires='>=3.8',
    package_data={
        'pogema_toolbox': ['maps/*.yaml'],
    },
)



from typing import Tuple
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from pogema_toolbox.views.view_plot import PlotView, prepare_plt, prepare_plot_fields
from pogema_toolbox.views.view_utils import eval_logs_to_pandas, drop_na

from typing import Literal


class MultiPlotView(PlotView):
    type: Literal['multi-plot'] = 'multi-plot'
    over: str = None
    num_cols: int = 3
    share_x: bool = False
    share_y: bool = False
    remove_individual_titles: bool = True
    legend_bbox_to_anchor: Tuple[float, float] = (0.5, -0.05)
    legend_loc: str = 'lower center'
    legend_columns: int = 5
    width: float = 2.0
    height: float = 2.0


def process_multi_plot_view(results, view: MultiPlotView, save_path=None):
    df = eval_logs_to_pandas(results)
    df = drop_na(df)
    if view.hue_order is None:
        view.hue_order = sorted(df['algorithm'].unique())

    if view.sort_by:
        df.sort_values(by=['map_name', 'algorithm'], inplace=True)

    if view.rename_fields:
        df = df.rename(columns=view.rename_fields)

    over_keys = sorted(df[view.over].unique())
    num_cols = view.num_cols
    num_rows = len(over_keys) // num_cols + (1 if len(over_keys) % num_cols else 0)

    prepare_plt(view)
    fig, axs = plt.subplots(num_rows, num_cols, figsize=(view.width * num_cols, view.height * num_rows),
                            sharex=view.share_x, sharey=view.share_y)

    # Adjust for when axs is not a 2D array
    if num_rows == 1 or num_cols == 1:
        axs = np.array(axs).reshape(num_rows, num_cols)

    x, y, hue = prepare_plot_fields(view)
    if view.ticks:
        plt.setp(axs, xticks=view.ticks)

    for idx, over in enumerate(over_keys):
        ax = axs[idx // num_cols, idx % num_cols]
        g = sns.lineplot(x=x, y=y, data=df[df[view.over] == over], errorbar=view.error_bar, hue=hue, ax=ax,
                         style=hue if view.line_types else None, markers=view.markers, palette=view.palette,
                         linewidth=view.line_width, hue_order=view.hue_order, style_order=view.hue_order)
        ax.set_title(over if not view.remove_individual_titles else '')

        if view.remove_individual_titles:
            legend = g.get_legend()
            if legend is not None:  # Check if the legend exists before removing
                legend.remove()

        if view.use_log_scale_x:
            ax.set_xscale('log', base=2)
            from matplotlib.ticker import ScalarFormatter
            ax.xaxis.set_major_formatter(ScalarFormatter())

        g.grid()

    # Remove unused axes
    for idx in range(len(over_keys), num_rows * num_cols):
        fig.delaxes(axs.flatten()[idx])

    if view.tight_layout:
        plt.tight_layout()

    # Handle legend outside the loop to prevent duplication
    handles, labels = ax.get_legend_handles_labels()
    if handles:
        fig.legend(handles, labels, bbox_to_anchor=view.legend_bbox_to_anchor, loc=view.legend_loc,
                   ncol=view.legend_columns, fancybox=True, shadow=False)

    if save_path:
        plt.savefig(save_path, bbox_inches='tight')
    plt.close(fig)
